(* algorithmic core *)
Require Import Ltac2.Ltac2.
Require UFProof.
Require UnionFind.
Require Helpers.

Ltac2 Type TermInfo := [ Atom | App(int, int) ].
Ltac2 Type ClassInfo := int list * int list.

Ltac2 Type rec 'a  CongrAxiom := [ Congr(int, int, int, int, ('a CongrAxiom) UFProof.t,  ('a CongrAxiom) UFProof.t) | Ax('a) ].
Ltac2 Type 'a Prf := ( 'a CongrAxiom) UFProof.t.

Ltac2 Type 'a State :=
  {
    termLookup : TermInfo array ;
    unionFind : ('a CongrAxiom, ClassInfo) UnionFind.t ;
    mutable equationsQue : (int * int * 'a Prf) list ;
    mutable congruencesQue : (int list * int list) list
  }.

Ltac2 classCombiner a b :=
    let (al, ar) := a in
    let (bl, br) := b in
      (List.append al bl, List.append ar br).

Ltac2 doEquation
  (cc: 'a State)
  (lhs: int)
  (rhs: int)
  (proof: 'a Prf) := 
  match (UnionFind.union classCombiner (cc.(unionFind)) lhs rhs proof) with
  | None => ()
  | Some x => let (ainfo,binfo) := x in
              let (af,ax) := ainfo in (* TOOD: rename these vars *)
              let (bf,bx) := binfo in
                 cc.(congruencesQue) := (af,bf) :: (ax,bx) :: cc.(congruencesQue)
  end.

Ltac2 doCongruence
  (cc: 'a State)
  (lhs: int)
  (rhs: int)
  := match (Array.get (cc.(termLookup)) lhs) with | Atom => ()
    | App lf lx => 
    match (Array.get (cc.(termLookup)) rhs) with | Atom => ()
    | App rf rx => 
    match (UnionFind.isCongruent (cc.(unionFind)) lhs rhs) with | true => ()
    | false =>
    match (UnionFind.congruent (cc.(unionFind)) lf rf) with | None => ()
    | Some fProof => 
    match (UnionFind.congruent (cc.(unionFind)) lx rx) with | None => ()
    | Some xProof => 
      doEquation cc lhs rhs (UFProof.axiom lhs rhs (Congr lf rf lx rx fProof xProof))
    end
    end
    end
    end
    end.

Ltac2 step (cc: 'a State) :=
  match (cc.(congruencesQue)) with
  | congr :: congrQue =>
     cc.(congruencesQue) := congrQue;
     let (lhss, rhss) := congr in
     List.iter (fun lhs => List.iter (doCongruence cc lhs) rhss) lhss;
     true
  | [] =>
     match (cc.(equationsQue)) with
     | e :: eqQue => 
       cc.(equationsQue) := eqQue;
       let (lhs, rhs, prf) := e in
       doEquation cc lhs rhs prf;
       true
     | [] => false
     end
  end.

Ltac2 rec stepUntillDone (cc: 'a State) :=
  match (step cc) with
  | true => stepUntillDone cc
  | false => ()
  end.

Ltac2 init (a : TermInfo array) (eqns: (int * int * 'a UFProof.t) list) :=
  let predLookups := Array.make (Array.length a) ( [], [] ) in
  Helpers.iterDown (Array.length a)
    (fun i => match (Array.get a i) with
              | Atom => ()
              | App lc rc =>
                    Helpers.update predLookups lc (fun (l, r) => (i :: l, r));
                    Helpers.update predLookups rc (fun (l, r) => (l,  i :: r))
              end);
  {
    termLookup := a ;
    unionFind := UnionFind.fromInfo predLookups ;
    congruencesQue := [] ;
    equationsQue := eqns
  }.

(*
  returns a union find structure of a congruence from the given term structure and equation list 
*)
Ltac2 findCongruence 
  (a: TermInfo array )
  (eqns:  (int * int * 'a CongrAxiom UFProof.t) list)
  := let cc := init a eqns in
     stepUntillDone cc;
     cc.(unionFind).
