Require Import Ltac2.Ltac2.
Require UFProof.
Require CongruenceClosure.
Require CCBuilder.

(* this is all the code that is eq specific, the last todo on a basic working verison is make this but with setoids *)

Lemma f_equal_split: forall {A B: Type} (f f': A -> B) (x x': A), f = f' -> x = x' -> f x = f' x'.
Proof.
  intros.
  now subst.
Defined.

Ltac2 rec buildEqProof (lookup: constr array) (prf: constr CongruenceClosure.Prf) 
  := match prf with
      | UFProof.Trans x h y k z 
        => let xt := Array.get lookup x in
            let yt := Array.get lookup y in
            let zt := Array.get lookup z in
            let l := buildEqProof lookup h in
            let r := buildEqProof lookup k in
            constr:(@eq_trans _ $xt $yt $zt $l $r)
      | UFProof.Sym x y h 
        => let xt := Array.get lookup x in
            let yt := Array.get lookup y in
            let l := buildEqProof lookup h in
            constr:(@eq_sym _ $xt $yt $l)
      | UFProof.Ax l r a0 => 
          match a0 with
          | CongruenceClosure.Ax a => a
          | CongruenceClosure.Congr f f' x x' h k => 
          let ft := Array.get lookup f in
          let ft' := Array.get lookup f' in
          let xt := Array.get lookup x in
          let xt' := Array.get lookup x' in
          let l := buildEqProof lookup h in
          let r := buildEqProof lookup k in
          constr:(@f_equal_split _ _ $ft $ft' $xt $xt' $l $r)
          end
      | UFProof.Refl x 
        => let xt := Array.get lookup x in
            constr:(@eq_refl _ $xt)
      end.

Ltac2 Type OverEqConcs := [ Goal ].

Ltac2 rec addTerm (builder : (constr, 'a, OverEqConcs) CCBuilder.t) (tm : constr)
      :=  CCBuilder.lookupOr builder (Constr.equal tm) 
            (fun () =>
                lazy_match! tm with
                | ?f ?x => lazy_match! (Constr.type f) with
                           | (_ -> _) => CCBuilder.makeApp builder tm (addTerm builder f) (addTerm builder x)
                           | _ => CCBuilder.makeAtom  builder tm
                           end
                | _ => CCBuilder.makeAtom  builder tm
                end).

Ltac2 rec addEquations (builder : (constr, constr, OverEqConcs) CCBuilder.t)
    := List.iter (fun (id,_,ty) => 
                      lazy_match! ty with
                      | ?a = ?b => CCBuilder.addEquation builder (addTerm builder a) (addTerm builder b) (CongruenceClosure.Ax (Control.hyp id) )
                      | _ => ()
                      end) 
        (Control.hyps ()).

Ltac2 add_goal (builder : (constr, constr, OverEqConcs) CCBuilder.t) :=
  lazy_match! (Control.goal ()) with
  | ?a = ?b => CCBuilder.addConclusion builder (addTerm builder a) (addTerm builder b) (Goal) 
  | _ => ()
  end.

Ltac2 run0 (u : unit) :=
  let builder := CCBuilder.new () in
  addEquations builder; add_goal builder;
  match (CCBuilder.run builder) with
  | CCBuilder.Sat _ _ => Control.zero (Tactic_failure (Some (Message.of_string "sat")))
  | CCBuilder.NoSat lookup lhs rhs proof conc 
      => let term := buildEqProof lookup proof in
         match conc with
         | Goal => Control.refine (fun () => term)
         end
  end.

Module Notations.
  Ltac2 Notation "basicCongruence" := Control.enter run0.
End Notations.
